import asyncio

import pytest

from action_trees.action_item import (
    ActionItem,
    ActionState,
    BlockingType,
    StateTransitionException,
    ActionFailedException,
    check_state_transition,
)

# pylint: disable=redefined-outer-name, protected-access


class ForcedException(Exception):
    """Test exception for testing"""


class ConcreteActionItem(ActionItem):
    """Concrete implementation of ActionItem for testing"""

    def __init__(self, name: str):
        super().__init__(name=name)
        self.initialized = False
        self.done = False

    async def _on_init(self):
        """initialize"""
        self._log.info("initializing")
        self.initialized = True

    async def _on_run(self):
        """run all children if any, otherwise just wait a bit"""
        try:
            if self.children:
                self._log.info("running children")
                for child in self.children:
                    await child.start()
            else:
                self._log.info("running self")
                for _ in range(5):
                    await self._wait_if_paused()
                    await asyncio.sleep(0.1)
        except ActionFailedException:
            if self.parent is not None:
                raise
            else:
                self._log.error("action failed")
                self.state = ActionState.FAILED
        except StateTransitionException:
            self._log.error("state transition failed")
            raise

        self.done = True


class SimpleActionItem(ActionItem):
    """action that does nothing, no init"""

    async def _on_run(self):
        pass


class FailingActionItem(ActionItem):
    """action that fails on run"""

    async def _on_run(self):
        raise ActionFailedException("Bummer!", self)


class BadActionItem(ActionItem):
    """action that fails on run"""

    async def _on_run(self):
        raise ForcedException("Failed without even trying")


def test_smoke():
    """basic smoke test"""
    state = ActionState.WAITING
    assert str(state) == "WAITING"

    blocking_type = BlockingType.HARD
    assert str(blocking_type) == "HARD"

    a1 = ConcreteActionItem(name="action1")
    a2 = ConcreteActionItem(name="action2")

    a1.add_child(a2)

    a1.display_tree()

    assert a1.get_exception() is None

    a1.state = a1.state


def test_transition_check():
    check_state_transition(ActionState.WAITING, ActionState.RUNNING)

    with pytest.raises(StateTransitionException):
        check_state_transition(ActionState.RUNNING, ActionState.PAUSED)

    with pytest.raises(StateTransitionException):
        check_state_transition(ActionState.WAITING, ActionState.FINISHED)


@pytest.fixture
def action_item():
    return ConcreteActionItem(name="action")


@pytest.mark.asyncio
async def test_initialization(action_item: ConcreteActionItem):
    assert action_item.state == ActionState.WAITING


@pytest.mark.asyncio
async def test_start(action_item: ConcreteActionItem):
    action_item.start()
    action_item.start()  # Should be a no-op
    await asyncio.sleep(0.01)  # Allow some time for the coroutine to start
    assert action_item.initialized
    assert action_item.state == ActionState.RUNNING
    await action_item.cancel()


@pytest.mark.asyncio
async def test_pause_and_resume(action_item: ConcreteActionItem):
    action_item.start()
    await asyncio.sleep(0.01)
    action_item.pause()
    assert action_item.state == ActionState.PAUSED
    await asyncio.sleep(0.1)
    action_item.resume()
    assert action_item.state == ActionState.RUNNING
    await asyncio.sleep(0.5)
    assert action_item.done
    assert action_item.state == ActionState.FINISHED
    assert action_item.get_exception() is None


@pytest.mark.asyncio
async def test_cancel(action_item: ConcreteActionItem):
    action_item.start()
    await asyncio.sleep(0.01)
    await action_item.cancel()
    await asyncio.sleep(0.01)  # Allow time for cancellation to propagate
    assert action_item.state == ActionState.FAILED
    await action_item.cancel()


@pytest.mark.asyncio
async def test_invalid_transition_pause(action_item: ConcreteActionItem):
    with pytest.raises(StateTransitionException):
        action_item.pause()


@pytest.mark.asyncio
async def test_invalid_transition_resume(action_item: ConcreteActionItem):
    with pytest.raises(StateTransitionException):
        action_item.resume()


@pytest.mark.asyncio
async def test_invalid_transition_cancel(action_item: ConcreteActionItem):
    action_item.start()
    await asyncio.sleep(0.01)
    action_item.pause()
    await asyncio.sleep(0.01)
    with pytest.raises(StateTransitionException):
        await action_item.cancel()

    action_item.resume()
    await asyncio.sleep(0.01)
    await action_item.cancel()

    exc = action_item.get_exception()
    assert isinstance(exc, asyncio.CancelledError)


@pytest.mark.asyncio
async def test_failed_action():
    action_item = BadActionItem(name="bad")

    action_item.start()
    await asyncio.sleep(0.01)

    assert action_item.state == ActionState.FAILED

    exc = action_item.get_exception()
    assert isinstance(exc, ForcedException)


def test_children():
    """test children"""
    action = ConcreteActionItem(name="parent")
    action.add_child(ConcreteActionItem(name="child1"))
    action.add_child(ConcreteActionItem(name="child2"))

    assert len(action.children) == 2

    assert action.get_child("child1").name == "child1"

    with pytest.raises(ValueError):
        action.get_child("child3")

    assert action.get_child("child2").name == "child2"


def test_parent():
    """test parent"""
    action = ConcreteActionItem(name="parent")
    child = ConcreteActionItem(name="child")
    action.add_child(child)

    assert child.parent == action
    action.remove_child("child")
    assert child.parent is None


@pytest.mark.asyncio
async def test_group_cancel():
    """failing child action must cancel its children and parent"""
    parent = ConcreteActionItem(name="parent")
    child = FailingActionItem(name="child")
    parent.add_child(child)
    grandchild = ConcreteActionItem(name="grandchild")
    child.add_child(grandchild)

    parent.start()
    await asyncio.sleep(0.1)

    assert parent.state == ActionState.FAILED
    assert child.state == ActionState.FAILED
    assert grandchild.state == ActionState.FAILED


@pytest.mark.asyncio
async def test_paused_cancel(action_item: ConcreteActionItem):
    """test cancelling a paused action"""

    child = SimpleActionItem(name="child")
    child._state = ActionState.PAUSED

    action_item.add_child(child)

    action_item.start()
    await asyncio.sleep(0.01)

    await action_item.cancel()
    await asyncio.sleep(0.01)
    exc = action_item.get_exception()
    assert isinstance(exc, StateTransitionException)

    assert child.state == ActionState.PAUSED
    assert isinstance(child.get_exception(), StateTransitionException)



#!/usr/bin/env python3
"""
 Base class for action items.

 Copyright (c) 2023 FILICS GmbH & ROX Automation
"""
from __future__ import annotations
from typing import Coroutine, Callable
import asyncio
import logging
from abc import ABC, abstractmethod
from enum import Enum, auto
from action_trees.protocols import ActionItemProtocol


class StateTransitionException(Exception):
    """Exception raised when an invalid state transition is attempted"""


class ActionFailedException(Exception):
    """Exception raised when an action can not be completed successfully"""

    def __init__(self, message: str, action: ActionItemProtocol):
        super().__init__(message)
        self.action = action

    def __str__(self):
        message = self.args[0] if self.args else ""
        return f"{message} in action '{self.action.name}' (class {self.action.__class__.__name__}))"


class BlockingType(Enum):
    """
    Enum representing the blocking types for actions in a VDA5050 compliant system.

    Attributes:
        NONE: Action can be executed in parallel with other actions and while the vehicle is driving.
        SOFT: Action can be executed in parallel with other actions, but the vehicle must not drive.
        HARD: Action must not be executed in parallel with other actions and the vehicle must not drive.
    """

    NONE = auto()
    SOFT = auto()
    HARD = auto()

    def __str__(self):
        return self.name


class ActionState(Enum):
    """Enum representing the states of an action item."""

    NONE = auto()
    WAITING = auto()
    INITIALIZING = auto()
    RUNNING = auto()
    PAUSED = auto()
    FINISHED = auto()
    FAILED = auto()

    def __str__(self):
        return self.name


# used for state transition checks
VALID_STATE_TRANSITIONS = [
    (ActionState.WAITING, ActionState.INITIALIZING),
    (ActionState.WAITING, ActionState.RUNNING),
    (ActionState.WAITING, ActionState.FAILED),
    (ActionState.INITIALIZING, ActionState.RUNNING),
    (ActionState.INITIALIZING, ActionState.FAILED),
    (ActionState.RUNNING, ActionState.FINISHED),
    (ActionState.RUNNING, ActionState.FAILED),
]


def check_state_transition(current_state: ActionState, new_state: ActionState):
    """Check if a state transition is valid.

    Args:
        current_state (ActionState): current state
        new_state (ActionState): new state

    Raises:
        StateTransitionException: if the transition is invalid
    """
    # switching between PAUSED and other states is only allowed by calling resume() or pause()
    if ActionState.PAUSED in (current_state, new_state):
        raise StateTransitionException(
            f"Invalid state transition: {current_state} -> {new_state}. You must call resume() or pause()"
        )

    # all other state transitions are checked against the VALID_STATE_TRANSITIONS list
    if (current_state, new_state) not in VALID_STATE_TRANSITIONS:
        raise StateTransitionException(
            f"Invalid state transition: {current_state} -> {new_state}"
        )


class ActionItem(ABC):
    """
    Abstract base class for action items.
    """

    # TODO: add callbacks for state changes

    def __init__(
        self,
        name: str,
        parent: ActionItemProtocol | None = None,
        logger: logging.Logger | None = None,
    ):
        """logger is optional, if not provided, a default logger will be used.
        using ROS logger is allowed."""
        self.name = name
        self.parent: ActionItemProtocol | None = parent or None
        self._children: dict[str, ActionItemProtocol] = {}

        self._log = logger or logging.getLogger(self.name)
        self._state: ActionState = ActionState.WAITING
        self._saved_state: ActionState = ActionState.NONE
        self._log.debug(f"initial state: {self._state}")

        # main task reference
        self._main_task: asyncio.Task | None = None

        # note: could use one event, but it would require polling waiting for resume
        self._pause_event = asyncio.Event()
        self._resume_event = asyncio.Event()

    def get_child(self, name: str) -> ActionItemProtocol:
        """get child by name"""
        try:
            return self._children[name]
        except KeyError as e:
            raise ValueError(f"no child with name {name}") from e

    def add_child(self, child: ActionItemProtocol) -> None:
        """add child"""
        self._children[child.name] = child
        child.parent = self

    def remove_child(self, name: str) -> None:
        """remove child by name"""
        self._children[name].parent = None
        del self._children[name]

    @property
    def children(self) -> list[ActionItemProtocol]:
        """get children"""
        return list(self._children.values())

    def display_tree(self, level=0):
        """
        Recursively display children and their states in a tree structure.
        """
        indent = "    " * level  # 4 spaces for each level of indentation
        state = self.state.name if self.state else "UNDEFINED"
        print(f"{indent}- {self.name} : {state}")
        for child in self.children:
            child.display_tree(level + 1)

    # ------------------ Actions ------------------
    def start(self) -> asyncio.Task:
        """Start the action item."""
        if self._main_task is None:
            self._main_task = asyncio.create_task(self.main())
        else:
            self._log.warning("action already started")

        return self._main_task

    async def main(self):
        """Main coroutine for the action item."""
        try:
            try:  # try to run the _on_init coroutine if it exists
                coro = getattr(self, "_on_init")
                await self._execute(coro, ActionState.INITIALIZING)
            except AttributeError:
                pass

            await self._execute(self._on_run, ActionState.RUNNING)

        except ActionFailedException as e:
            self._log.error(f"Failed. {e}")
            raise
        except StateTransitionException as e:
            self._log.error(f"State transition failed. {e}")
            raise

        except Exception as e:
            self._log.exception(f"Error occurred during execution: {e}")
            raise
        finally:
            self._log.info(f"Action {self.name} finished with state {self._state}")

    async def cancel(self):
        """Cancel the action item asynchronously."""

        self._log.info(f"canceling action {self.name}")

        # if the action is already finished, do nothing
        if self._state in [
            ActionState.FINISHED,
            ActionState.FAILED,
        ]:
            self._log.debug(
                f"action {self.name} is in {self._state} state, skipping cancel"
            )
            return

        # cancelling from PAUSED is not allowed
        check_state_transition(self._state, ActionState.FAILED)

        # cancel children
        await self._cancel_children()

        # cancel main task
        if self._main_task is not None:
            try:
                if self._main_task.cancel():
                    # Wait for the task to be cancelled
                    await self._main_task
            except asyncio.CancelledError:
                self._log.debug("task cancelled")
                # don't re-raise here, the task is cancelled by command.

        self.state = ActionState.FAILED

    def pause(self):
        """Pause the action item."""
        self._log.debug(f"Pausing {self.name}")
        if self._state not in [ActionState.INITIALIZING, ActionState.RUNNING]:
            raise StateTransitionException(
                f"Cannot pause from {self._state}, must be INITIALIZING or RUNNING"
            )
        self._pause_event.set()
        self._resume_event.clear()
        self._saved_state = self._state
        self._state = ActionState.PAUSED

    def resume(self):
        """Resume the action item."""
        self._log.debug(f"Resuming {self.name}")
        if self._state != ActionState.PAUSED:
            raise StateTransitionException(
                f"Cannot resume from {self._state}, must be PAUSED"
            )

        self._pause_event.clear()
        self._resume_event.set()
        self._state = self._saved_state
        self._saved_state = ActionState.NONE

    @property
    def state(self) -> ActionState:
        """Get the state of the action item."""
        return self._state

    @state.setter
    def state(self, state: ActionState) -> None:
        """Set the state of the action item."""
        if state == self._state:
            return

        check_state_transition(self._state, state)

        self._log.debug(f"Changing state {self._state} -> {state}")
        self._state = state

    def get_exception(self) -> BaseException | None:
        """Get the exception raised by the action item."""

        if self._main_task is None or not self._main_task.done():
            return None

        try:
            return self._main_task.exception()
        except asyncio.CancelledError:
            return asyncio.CancelledError(f"Action {self.name} was cancelled")

    async def _wait_if_paused(self):
        """Wait until the action is resumed"""
        if self._pause_event.is_set():
            self._log.debug("paused...")
            await self._resume_event.wait()
            self._log.debug("resumed")

    async def _cancel_children(self):
        """Cancel all children of this action item"""

        self._log.debug(f"Canceling children of {self.name}")
        for child in self.children:
            self._log.debug(f"Canceling child {child.name}")
            await child.cancel()

    async def _execute(
        self, action: Callable[[], Coroutine], action_state: ActionState
    ) -> None:
        """execute an action coroutine and handle exceptions

        Args:
            action (Callable[[], Coroutine]): coroutine to execute
            action_state (ActionState): state to set while executing the coroutine

        Raises:
            exc: exception raised by the coroutine
        """

        self._log.debug(f"Executing {action.__name__}")
        self.state = action_state

        # just an extra check to make sure the events are not set
        assert not self._pause_event.is_set(), "pause event should not be set"
        assert not self._resume_event.is_set(), "resume event should not be set"

        try:
            await action()
        except asyncio.CancelledError:
            self._log.debug(f"action {self.name} was cancelled")
            self.state = ActionState.FAILED
            raise
        except ActionFailedException:
            self.state = ActionState.FAILED
            # cancel children
            await self._cancel_children()
            raise

        except Exception as exc:
            self._log.error(f"Error in {action.__name__}: {exc}")
            self.state = ActionState.FAILED
            raise exc

        if self._state == ActionState.RUNNING:
            self.state = ActionState.FINISHED

    # ------------------ Action coroutines ------------------
    # This is where the actual work is done.
    # They are meant to be overridden by the user.

    @abstractmethod
    async def _on_run(self):
        """main coroutine for running the action item."""



""" action protocols """
from __future__ import annotations
import asyncio
from typing import Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from .action_item import ActionState


class ActionItemProtocol(Protocol):
    """interface for action items"""

    def start(self) -> asyncio.Task:
        """start action item, if not already started, always return reference
        to main task"""

    async def cancel(self) -> None:
        """asynchronously cancel the action item"""

    def pause(self) -> None:
        ...

    def resume(self) -> None:
        ...

    def get_exception(self) -> BaseException | None:
        """get exception if any"""

    @property
    def state(self) -> ActionState:
        ...

    @property
    def name(self) -> str:
        ...

    @property
    def parent(self) -> ActionItemProtocol | None:
        """get parent (if any)"""

    @parent.setter
    def parent(self, parent: ActionItemProtocol | None) -> None:
        """set parent"""

    def add_child(self, child: ActionItemProtocol) -> None:
        """add child"""

    def get_child(self, name: str) -> ActionItemProtocol:
        """get child by name"""

    def remove_child(self, name: str) -> None:
        """remove child by name"""

    @property
    def children(self) -> list[ActionItemProtocol]:
        """get children"""



#!/usr/bin/env python3
"""
Action tree example - actions with children

Explained with an example of a coffee making machine.

Action tree, items with "&" after them are parallel to following item.

- cappuccino order
    - prepare machine
        - intialize
        - clean
    - make cappuccino
        - boil water &
        - grind coffee
        - turn on pump for 2 seconds
        - add milk
    - goto standby

Note: for simplicity, pausing actions is not implemented here.

Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging
import coloredlogs

import action_trees
from action_trees import ActionItem, ActionState, ActionFailedException
from action_trees.protocols import ActionItemProtocol


class AtomicAction(ActionItem):
    """basic machine action. These actions have no children and are atomic"""

    def __init__(self, name: str, duration: float):
        super().__init__(name=name)
        self._duration = duration

    async def _on_run(self):
        """Run the action item."""
        self._log.info(f"Running {self.name}")
        try:
            await asyncio.sleep(self._duration)
        except asyncio.CancelledError:
            self._log.info(f"{self.name} running cancel tasks...")
            raise


class FailingGrindAction(AtomicAction):
    """crashing version of AtomicAction"""

    async def _on_run(self):
        """Run the action item."""
        self._log.info(f"Running {self.name}")
        try:
            raise ActionFailedException("No beans left!", self)
        except asyncio.CancelledError:
            self._log.warning(f"{self.name} cancelled")
            raise


class PrepareMachineAction(ActionItem):
    """prepare the machine"""

    def __init__(self):
        super().__init__(name="prepare")

    async def _on_init(self):
        self._log.info("initializing")
        await asyncio.sleep(1)

    async def _on_run(self):
        try:
            self._log.info("cleaning")
            await asyncio.sleep(1)
            self._log.info("cleaning done")
        except asyncio.CancelledError:
            self._log.warning("cancelling cleaning")
            self._log.info("performing cancel actions...")
            await asyncio.sleep(1)
            raise


class MakeCappuccinoAction(ActionItem):
    """make cappuccino"""

    def __init__(self):
        super().__init__(name="make_cappuccino")

        # add children
        self.add_child(AtomicAction(name="boil_water", duration=2))
        self.add_child(AtomicAction(name="grind_coffee", duration=1))
        self.add_child(AtomicAction(name="pump", duration=2))
        self.add_child(AtomicAction(name="add_milk", duration=1))

    async def _on_run(self):
        # try:
        self._log.info("making cappuccino")
        # parallel actions
        await asyncio.gather(
            self.get_child("boil_water").start(),
            self.get_child("grind_coffee").start(),
        )

        # sequential actions
        await self.get_child("pump").start()
        await self.get_child("add_milk").start()
        self._log.info("making cappuccino done")


class CappuccinoOrder(ActionItem):
    """make cappuccino, highest level action"""

    def __init__(self):
        super().__init__(name="cappuccino order")

        # add children
        self.add_child(PrepareMachineAction())
        self.add_child(MakeCappuccinoAction())
        self.add_child(AtomicAction(name="goto_standby", duration=1))

    async def _on_run(self):
        self._log.info("starting cappuccino order")
        try:
            for action_name in ["prepare", "make_cappuccino", "goto_standby"]:
                await self.get_child(action_name).start()

            self._log.info("Cappuccino is ready!")
        except ActionFailedException as e:
            # explicitly handle ActionFailedException here, as it is the higest level.
            # we can choose to how to cleanup here, like canceling children actions.
            self._log.error(f"Sorry, no cappuccino today: {e}")
            await self._cancel_children()
            self.state = ActionState.FAILED


async def main() -> None:
    """main coroutine"""

    logging.info("-------------------Example 1: normal execution------------------")
    order = CappuccinoOrder()
    await order.start()
    order.display_tree()

    logging.info("-------------------Example 2: one of actions fails------------------")
    order = CappuccinoOrder()

    # add a bad action  - grind_coffee will fail.
    action = order.get_child("make_cappuccino")
    action.remove_child("grind_coffee")
    action.add_child(FailingGrindAction(name="grind_coffee", duration=1))
    try:
        await order.start()
    except asyncio.CancelledError:
        logging.info("Cancelled")
    logging.info(f"order state: {order.state}")
    order.display_tree()

    logging.info("-------------------Example 3: cancelling------------------")
    order = CappuccinoOrder()
    order.start()
    await asyncio.sleep(3)
    logging.warning("cancelling order")
    await order.cancel()
    logging.info(f"order state: {order.state}")
    order.display_tree()


if __name__ == "__main__":
    coloredlogs.install(
        level="DEBUG",
        fmt=action_trees.LOG_FORMAT,
    )

    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("interrupted")



#!/usr/bin/env python3
"""
action_trees CLI
"""

import click
from action_trees import __version__


@click.group()
def cli():
    pass  # pragma: no cover


@cli.command()
def info():
    """ Print package info """
    print(__version__)


cli.add_command(info)

if __name__ == "__main__":
    cli()  # pragma: no cover



from .version import __version__
from .action_item import ActionItem, ActionState, ActionFailedException


LOG_FORMAT = "%(asctime)s,%(msecs)03d %(name)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s"



# type: ignore

"""The setup script."""


import codecs
import os
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


requirements = ["click"]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Jev Kuznetsov",
    author_email="jev@roxautomation.com",
    setup_requires=["setuptools_scm"],
    python_requires=">=3.11",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
    ],
    description="Action decomposition and execution framework",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    install_requires=requirements,
    include_package_data=True,
    keywords="",
    name="action_trees",
    package_dir={"": "src"},
    packages=find_packages("src"),
    test_suite="tests",
    tests_require=test_requirements,
    url="",
    # Use setuptools_scm for versioning
    use_scm_version={
        "write_to": "src/action_trees/version.py",
        "fallback_version": "0.0.0+unknown",
    },
    # zip_safe=False,
    entry_points={"console_scripts": ["action_trees=action_trees.cli:cli"]},
)



#!/usr/bin/env python3
"""
 Basic action item example

 Copyright (c) 2024 ROX Automation - Jev Kuznetsov
"""

import asyncio
import logging
import coloredlogs

from action_trees import ActionItem, ActionState, ActionFailedException
from action_trees.protocols import ActionItemProtocol
from action_trees import LOG_FORMAT


class MyFirstAction(ActionItem):
    """Example of an action item."""

    def __init__(self):
        super().__init__(name="action1")

    async def _on_init(self):
        """Initialize the action item."""
        self._log.info("Starting init")
        try:
            for idx in range(3):
                self._log.info(f"init {idx}")
                await self._wait_if_paused()
                await asyncio.sleep(0.1)
            self._log.info("Finished init")
        except asyncio.CancelledError:
            self._log.warning("on_init cancelled")
            # do some cleanup here
            raise

    async def _on_run(self):
        """Run the action item."""
        self._log.info("Running")
        idx = 0
        try:
            while True:
                self._log.info(f"run {idx}")
                await self._wait_if_paused()
                idx += 1
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            self._log.warning("on_run cancelled")
            raise


class InstantAction(ActionItem):
    """action item that runs instantly"""

    def __init__(self):
        super().__init__(name="action2")
        self.start()

    async def _on_run(self):
        """Run the action item."""
        self._log.info("Running")
        idx = 0
        try:
            for idx in range(3):
                self._log.info(f"run {idx}")
                await self._wait_if_paused()
                idx += 1
                await asyncio.sleep(1)
            self._log.info("Finished")
        except asyncio.CancelledError:
            self._log.warning("on_run cancelled")
            raise


async def command_action(action: ActionItemProtocol):
    """send some actions to the action item"""
    await asyncio.sleep(2)
    action.pause()
    await asyncio.sleep(2)
    action.resume()
    await asyncio.sleep(3)
    await action.cancel()
    await asyncio.sleep(2)
    logging.info("Done.")


async def main() -> None:
    """main coroutine"""

    action1: ActionItemProtocol = MyFirstAction()
    tsk1 = action1.start()

    action2: ActionItemProtocol = InstantAction()
    tsk2 = action2.start()  # noop - already started

    await command_action(action1)

    # check task status
    logging.info(f"{tsk1.done()=}, {tsk2.done()=}")

    # get exception if any
    for action in (action1, action2):
        exc = action.get_exception()
        logging.info(f"action {action.name} {action.state}  exception: {exc}")


if __name__ == "__main__":
    coloredlogs.install(
        level="DEBUG",
        fmt=LOG_FORMAT,
    )

    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("interrupted")
